==> blobAndConquer-0.1-1.tar.gz: BlobAndConquer/src/units.cpp <==
/*
Copyright (C) 2005 Parallel Realities

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "units.h"

void throwUnit(Unit *unit, Vector velocity)
{
	if (unit->helpless == 0)
	{
		unit->helpless = 200;
	}
	
	unit->velocity = velocity;
	unit->velocity.x /= 2;
	unit->velocity.y /= 2;
	unit->velocity.z = 0.5;
	
	Math::addBit(&unit->flags, EF_BOUNCES);
}

void updateUnitContentTypes(Unit *unit)
{
	
}

void unitContentChanged(Unit *unit)
{
	Vector position;
	int newContentType;
	int envContents[3];
	
	envContents[0] = envContents[1] = envContents[2] = CONTENTS_SOLID;
	
	unit->liquidLevel = 0;
	
	// Check content types... going from head to feet ;)
	// Should look at this really since it's probably bloody slow... :|
	
	position = unit->position;
	position.z += unit->boundingBox.maxs.z;
	envContents[0] = bsp->getContentsAtPosition(position);
	
	position = unit->position;
	envContents[1] = bsp->getContentsAtPosition(position);
	
	position = unit->position;
	position.z += unit->boundingBox.mins.z;
	envContents[2] = bsp->getContentsAtPosition(position);
	
	newContentType = tracer->outputContent;
	
	if (envContents[1] & CONTENTS_WATER)
	{
		unit->liquidLevel = 1;
	}
	
	if (envContents[0] & CONTENTS_WATER)
	{
		unit->liquidLevel = 2;
		//unit->oxygen = min(unit->oxygen, MAX_OXYGEN);
	}
	
	if (newContentType == unit->contentType)
	{
		return;
	}
	
	// Unit is moving in to liquid
	if (newContentType & (CONTENTS_AIR|CONTENTS_WATER|CONTENTS_SLIME|CONTENTS_LAVA))
	{
		if (unit->contentType & (CONTENTS_AIR|CONTENTS_SOLID))
		{
			Math::limit(&unit->velocity.z, -0.5, 0.5);
			
			if (tracer->outputContent & CONTENTS_WATER)
			{
				audio->playSound(SND_WATER_IN, (unit == player) ? CH_PLAYER : CH_ANY);
			}
			else if (tracer->outputContent & CONTENTS_SLIME)
			{
				audio->playSound(SND_SLIME_IN, (unit == player) ? CH_PLAYER : CH_ANY);
			}
			else if (tracer->outputContent & CONTENTS_LAVA)
			{
				audio->playSound(SND_LAVA_IN, (unit == player) ? CH_PLAYER : CH_ANY);
			}
		}
	}
	
	// Unit is moving out of water
	if (!(unit->flags & EF_DYING))
	{
		if (newContentType & (CONTENTS_AIR|CONTENTS_SOLID))
		{
			if (unit->contentType & CONTENTS_WATER)
			{
				if ((unit->velocity.z > 0) && (unit->helpless == 0))
				{
					unit->velocity.z = 1.25;
				}
				
				audio->playSound(SND_WATER_OUT, (unit == player) ? CH_PLAYER : CH_ANY);
			}
		}
	}
	
	unit->contentType = newContentType;
}

void keepUnitAboveLiquid(Unit *unit)
{
	if (unit->velocity.z >= 0)
	{
		return;
	}
	
	Vector start = unit->position;
	Vector end = start;
	
	end.z -= 15;
	
	bsp->traceRay(start, end);
	
	if (!(tracer->outputContent & CONTENTS_SOLID))
	{
		unit->velocity.z = 0;
	}
}

bool unitFallSafe(Unit *unit, Vector traceLocation)
{
	static int contents;
	static int range;
	
	range = 7;
	
	if (unit != NULL)
	{
		range = (unit->entityType == ET_BLOB) ? 250 : 7;
	}
	
	for (int i = 6 ; i < range ; i += 6)
	{
		traceLocation.z -= i;
		
		contents = bsp->getContentsAtPosition(traceLocation);
		
		if (contents & (CONTENTS_SLIME|CONTENTS_LAVA|CONTENTS_WATER))
		{
			return false;
		}
	}
	
	return true;
}

bool unitCanSpawnInLocation(Unit *unit, Vector location)
{
	static Vector start, end;
	
	start = location;
	
	end = start;
	
	if (unit != NULL)
	{
		end.z -= (unit->entityType == ET_BLOB) ? 250 : 6;
	}
	else
	{
		end.z -= 250;
	}
	
	if (!bsp->canSightTarget(start, end))
	{
		return unitFallSafe(unit, start);
	}
	
	return bbManager->canSightSolid(unit, start, end);
}

bool unitIsAtEdge(Unit *unit)
{
	static float r, angle, dx, dy;
	static Vector v, start, end, target;
	
	target = unit->position;
	target += unit->velocity;
	
	dx = unit->position.x - target.x;
	dy = unit->position.y - target.y;
	 
	angle = atan2(dy, dx);
	angle = -90 + toDegrees(angle);
	
	r = toRadians(angle);
	
	v = unit->position;
	v.x += (sin(r) * 3);
	v.y -= (cos(r) * 3);
	v -= unit->position;
	v.normalize();
	
	start = unit->position;
	start += (v * 3);
	
	return (!unitCanSpawnInLocation(unit, start));
}

void keepUnitAwayFromEdges(Unit *unit)
{
	if ((unit->velocity.x == 0) && (unit->velocity.y == 0))
	{
		return;
	}
	
	if (unit->helpless > 0)
	{
		return;
	}
	
	if (unit->flags & EF_WEIGHTLESS)
	{
		keepUnitAboveLiquid(unit);
		return;
	}
	
	if (!(unit->flags & EF_ONGROUND))
	{
		return;
	}
	
	if (unitIsAtEdge(unit))
	{
		unit->velocity.x = 0;
		unit->velocity.y = 0;
		unit->thinkTime = 0;
	}
}

bool unitCanSeeTarget(Entity *observer, Entity *target, Vector startPosition)
{
	if (!bsp->canSightTarget(startPosition, target->position))
	{
		return false;
	}
			
	if (!bbManager->canSightTarget(observer, target))
	{
		return false;
	}
	
	return true;
}

void teleportUnit()
{
	self->custom -= engine->getTimeDifference(TD_LOGIC);
	
	if (self->custom <= 0)
	{
		Math::removeBit(&self->flags, EF_TELEPORTING);
		self->position = self->destination;
		addTeleportParticles(self->position);
		audio->playSound(SND_TELEPORT3, (self == player) ? CH_PLAYER : CH_SPAWN);
		
		if (self != player)
		{
			self->action = self->walk;
		}
		else
		{
			self->action = NULL;
		}
	}
}

void faceLocation(Unit *unit, Vector location)
{
	float dx = unit->position.x - location.x;
	float dy = unit->position.y - location.y;
	float dz = unit->position.z - location.z;
	
	float angle;
	 
	angle = atan2(dy, dx);
	unit->rotation.x = -90 + toDegrees(angle);
	
	//angle = atan2(dx, dz);
	//unit->rotation.z = toDegrees(angle);
}

void faceTarget(Unit *unit)
{
	if (unit->target == NULL)
	{
		return;
	}
	
	faceLocation(unit, unit->target->position);
}

void doShieldHit(Unit *unit)
{
	if (unit->shieldHit)
	{
		glPushMatrix();
		{
			glTranslatef(unit->position.x, unit->position.y, unit->position.z);
			
			glDisable(GL_TEXTURE_2D);
			glEnable(GL_BLEND);
			
			GLColor c = GLColor::blue;
			c.color[3] = unit->shieldHit;
			glColor4fv(c.color);
			
			gluSphere(graphics->quadric, 4, 32, 16);
			
			glEnable(GL_TEXTURE_2D);
			glDisable(GL_BLEND);
		}
		glPopMatrix();
		
		unit->shieldHit -= (0.01 * engine->getTimeDifference(TD_LOGIC));
		Math::limit(&unit->shieldHit, 0, 2.0);
	}
}

void drawDroid(Unit *unit)
{
	glEnable(GL_TEXTURE_GEN_S);
	glEnable(GL_TEXTURE_GEN_T);
	
	glPushMatrix();
	{
		glTranslatef(unit->position.x, unit->position.y, unit->position.z);
		
		if (unit->helpless)
		{
			unit->rotation.x += (3 * engine->getTimeDifference(TD_LOGIC));
			unit->rotation.y += (3 * engine->getTimeDifference(TD_LOGIC));
		}
		
		glRotatef(unit->rotation.x, 0.0, 0.0, 1.0);
		glRotatef(unit->rotation.y, 0.0, 1.0, 0.0);
		
		if (unit->definition->model != NULL)
		{
			glCallList(unit->definition->model->data);
		}
	}
	glPopMatrix();
	
	glDisable(GL_TEXTURE_GEN_S);
	glDisable(GL_TEXTURE_GEN_T);
	
	addGravUnitParticles(unit);
}

void drawBlob(Unit *unit)
{
	glPushMatrix();
	{
		glTranslatef(unit->position.x, unit->position.y, unit->position.z);
		
		if (unit->helpless)
		{
			unit->rotation.x += (3 * engine->getTimeDifference(TD_LOGIC));
			unit->rotation.y += (3 * engine->getTimeDifference(TD_LOGIC));
		}
		
		glRotatef(unit->rotation.x, 0.0, 0.0, 1.0);
		glRotatef(unit->rotation.y, 0.0, 1.0, 0.0);
		//glRotatef(unit->rotation.z, 1.0, 0.0, 0.0);
		
		if (unit->definition->model != NULL)
		{
			glCallList(unit->definition->model->data);
		}
	}
	glPopMatrix();
}

void drawUnits()
{
	glDisable(GL_BLEND);
	glEnable(GL_TEXTURE_2D);
	glEnable(GL_LIGHTING);
	
	for (Unit *unit = (Unit*)entityManager->blobList.getFirstElement() ; unit != NULL ; unit = (Unit*)unit->next)
	{
		if ((unit->flags & EF_DEAD) || (unit->flags & EF_TELEPORTING) || (unit->flags & EF_VANISHED))
		{
			continue;
		}
		
		if (!withinViewableRange(unit->position, 6))
		{
			continue;
		}
		
		glBindTexture(GL_TEXTURE_2D, unit->definition->texture->data);
		
		drawBlob(unit);
		doShieldHit(unit);
		
		//drawBoundingBox(unit);
	}
	
	for (Unit *unit = (Unit*)entityManager->enemyList.getFirstElement() ; unit != NULL ; unit = (Unit*)unit->next)
	{
		if ((unit->flags & EF_DEAD) || (unit->flags & EF_TELEPORTING) || (unit->flags & EF_VANISHED))
		{
			continue;
		}
		
		if (!withinViewableRange(unit->position, 6))
		{
			continue;
		}
		
		glBindTexture(GL_TEXTURE_2D, unit->definition->texture->data);
			
		if (!(unit->flags & EF_WEIGHTLESS))
		{
			drawBlob(unit);
		}
		else
		{
			drawDroid(unit);
		}
		
		doShieldHit(unit);
		
		//drawBoundingBox(unit);
	}
	
	glDisable(GL_LIGHTING);
}

char *getUnitStatus(Unit *unit)
{
	float healthPercent = unit->health;
	healthPercent /= unit->maxHealth;
	healthPercent *= 100;
	
	if (healthPercent <= 0)
	{
		return localeManager->translate("STATUS_KIA");
	}
	else if (healthPercent <= 50)
	{
		return localeManager->translate("STATUS_WOUNDED");
	}
	else
	{
		return localeManager->translate("STATUS_OK");
	}
}

==> blobAndConquer-0.2-1.tar.gz: BlobAndConquer/src/units.cpp <==
/*
Copyright (C) 2005 Parallel Realities

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "units.h"

void drawUnit(bool transparent)
{
	Unit *unit = (Unit*)self;
	
	glDisable(GL_BLEND);
	glEnable(GL_TEXTURE_2D);
	glEnable(GL_LIGHTING);
	
	if (!transparent)
	{
		glBindTexture(GL_TEXTURE_2D, self->definition->texture->data);
		
		glPushMatrix();
		{
			glTranslatef(unit->position.x, unit->position.y, unit->position.z);
			
			if (unit->helpless)
			{
				unit->rotation.x += (3 * engine->getTimeDifference(TD_LOGIC));
				unit->rotation.y += (3 * engine->getTimeDifference(TD_LOGIC));
			}
			
			glRotatef(unit->rotation.x, 0.0, 0.0, 1.0);
			glRotatef(unit->rotation.y, 0.0, 1.0, 0.0);
			
			if (unit->definition->model != NULL)
			{
				glCallList(unit->definition->model->data);
			}
		}
		glPopMatrix();
	}
	else
	{
		if (unit->shieldHit)
		{
			glPushMatrix();
			{
				glTranslatef(unit->position.x, unit->position.y, unit->position.z);
			
				glDisable(GL_TEXTURE_2D);
				glEnable(GL_BLEND);
			
				GLColor c = GLColor::blue;
				c.color[3] = unit->shieldHit;
				glColor4fv(c.color);
			
				gluSphere(graphics->quadric, 4, 32, 16);
			
				glEnable(GL_TEXTURE_2D);
				glDisable(GL_BLEND);
			}
			glPopMatrix();
		
			unit->shieldHit -= (0.01 * engine->getTimeDifference(TD_LOGIC));
			Math::limit(&unit->shieldHit, 0, 2.0);
		}
		
		if (unit->flags & EF_FROZEN)
		{
			glEnable(GL_TEXTURE_2D);
			glEnable(GL_BLEND);
			
			glBindTexture(GL_TEXTURE_2D, textureManager->getTexture("IceBlock")->data);
			
			float size = unit->boundingBox.maxs.x * 2;
			
			glColor3f(0.5, 0.5, 0.5);
			drawBox(unit->position, size, size, size);
		}
	}
}

void throwUnit(Unit *unit, Vector velocity)
{
	if ((unit == player) && (game->cheatStun))
	{
		return;
	}
	
	if (unit->helpless == 0)
	{
		unit->helpless = 200;
	}
	
	unit->velocity = velocity;
	unit->velocity.x /= 2;
	unit->velocity.y /= 2;
	unit->velocity.z = 0.5;
	
	Math::addBit(&unit->flags, EF_BOUNCES);
}

void unitContentChanged(Unit *unit)
{
	Vector position;
	int newContentType;
	int envContents[3];
	
	envContents[0] = envContents[1] = envContents[2] = CONTENTS_SOLID;
	
	unit->liquidLevel = 0;
	
	// Check content types... going from head to feet ;)
	// Should look at this really since it's probably bloody slow... :|
	
	position = unit->position;
	position.z += unit->boundingBox.maxs.z;
	envContents[0] = bsp->getContentsAtPosition(position);
	
	position = unit->position;
	envContents[1] = bsp->getContentsAtPosition(position);
	
	position = unit->position;
	position.z += unit->boundingBox.mins.z;
	envContents[2] = bsp->getContentsAtPosition(position);
	
	newContentType = tracer->outputContent;
	
	if (envContents[1] & CONTENTS_WATER)
	{
		unit->liquidLevel = 1;
	}
	
	if (envContents[0] & (CONTENTS_WATER|CONTENTS_SLIME|CONTENTS_LAVA))
	{
		unit->liquidLevel = 2;
		unit->oxygen = min(unit->oxygen, MAX_OXYGEN);
	}
	
	if (newContentType == unit->contentType)
	{
		return;
	}
	
	// Unit is moving in to liquid
	if (newContentType & (CONTENTS_AIR|CONTENTS_WATER|CONTENTS_SLIME|CONTENTS_LAVA))
	{
		if (unit->contentType & (CONTENTS_AIR|CONTENTS_SOLID))
		{
			Math::limit(&unit->velocity.z, -0.5, 0.5);
			
			if (tracer->outputContent & CONTENTS_WATER)
			{
				audio->playSound(SND_WATER_IN, (unit == player) ? CH_PLAYER : CH_ANY);
			}
			else if (tracer->outputContent & CONTENTS_SLIME)
			{
				audio->playSound(SND_SLIME_IN, (unit == player) ? CH_PLAYER : CH_ANY);
			}
			else if (tracer->outputContent & CONTENTS_LAVA)
			{
				audio->playSound(SND_LAVA_IN, (unit == player) ? CH_PLAYER : CH_ANY);
			}
		}
	}
	
	// Unit is moving out of water
	if (!(unit->flags & EF_DYING))
	{
		if (newContentType & (CONTENTS_AIR|CONTENTS_SOLID))
		{
			if (unit->contentType & CONTENTS_WATER)
			{
				if ((unit->velocity.z > 0) && (unit->helpless == 0))
				{
					unit->velocity.z = 1.25;
				}
				
				audio->playSound(SND_WATER_OUT, (unit == player) ? CH_PLAYER : CH_ANY);
			}
		}
	}
	
	unit->contentType = newContentType;
}

void keepUnitAboveLiquid(Unit *unit)
{
	if (unit->velocity.z >= 0)
	{
		return;
	}
	
	Vector start = unit->position;
	Vector end = start;
	
	end.z -= 15;
	
	bsp->traceRay(start, end);
	
	if (!(tracer->outputContent & CONTENTS_SOLID))
	{
		unit->velocity.z = 0;
	}
}

void keepUnitInLiquid(Unit *unit)
{
	if (unit->velocity.z <= 0)
	{
		return;
	}
	
	Vector start = unit->position;
	Vector end = start;
	
	end.z += 15;
	
	bsp->traceRay(start, end);
	
	if (!(tracer->outputContent & CONTENTS_WATER))
	{
		unit->velocity.z = 0;
	}
}

bool unitCanSpawnInLocation(Unit *unit, Vector start)
{
	Vector end = start;
	
	if (unit != NULL)
	{
		end.z -= (unit->entityType == ET_BLOB) ? 100 : 7;
	}
	else
	{
		end.z -= 100;
	}
	
	bsp->traceRay(start, end);
	
	if (tracer->outputContent & (CONTENTS_WATER|CONTENTS_SLIME|CONTENTS_LAVA))
	{
		return false;
	}
	
	if (tracer->outputFraction == 1.0)
	{
		return bbManager->canSightSolid(unit, start, end);
	}
	
	return true;
}

Vector getUnitForwardVector(Entity *unit)
{
	Vector v;
	
	float r = toRadians(unit->rotation.x);
	
	v = unit->position;
	
	v.x += (sin(r) * 25);
	v.y -= (cos(r) * 25);
	v -= unit->position;
	v.normalize();
	
	return v;
}

bool unitIsAtEdge(Unit *unit)
{
	static float r, angle, dx, dy;
	static Vector v, start, end, target;
	
	target = unit->position;
	target += unit->velocity;
	
	dx = unit->position.x - target.x;
	dy = unit->position.y - target.y;
	 
	angle = atan2(dy, dx);
	angle = -90 + toDegrees(angle);
	
	r = toRadians(angle);
	
	v = unit->position;
	v.x += (sin(r) * 3);
	v.y -= (cos(r) * 3);
	v -= unit->position;
	v.normalize();
	
	start = unit->position;
	start += (v * 4);
	
	end = start;
	end.z += unit->boundingBox.mins.z;
	
	// don't allow Blobs to walk over edges if they
	// are in combat...
	if (unit->entityType == ET_BLOB)
	{
		if (unit->definition->type == BLOB_MIA)
		{
			end.z--;
		}
		else if (unit->target == NULL)
		{
			//end.z -= 185; // approximately how fall you can fall without becoming helpless
			end.z -= 32;
		}
		else
		{
			end.z--;
		}
	}
	else
	{
		// enemies can't walk off edges when they have no target...
		if (unit->target == NULL)
		{
			end.z--;
		}
		else
		{
			end.z -= 12;
		}
	}
	
	tracer->treatAllAsSolid = true;
	
	bsp->traceRay(start, end);
	
	if (tracer->outputContent & (CONTENTS_SLIME|CONTENTS_LAVA))
	{
		return true;
	}
	
	if (tracer->outputFraction == 1.0)
	{
		if (tracer->outputContent & (CONTENTS_WATER))
		{
			return false;
		}
		
		return (!bbManager->canSightSolid(unit, start, end));
	}
	
	return false;
}

void keepUnitAwayFromEdges(Unit *unit)
{
	//debug(("%s\n", unit->position.toString()));
	
	if (unit->helpless > 0)
	{
		return;
	}
	
	if (unit->flags & EF_WEIGHTLESS)
	{
		keepUnitAboveLiquid(unit);
		return;
	}
	
	if (unit->flags & EF_SWIMS)
	{
		keepUnitInLiquid(unit);
		return;
	}
	
	if ((unit->velocity.x == 0) && (unit->velocity.y == 0))
	{
		return;
	}
	
	if (!(unit->flags & EF_ONGROUND))
	{
		// check we're not close enough to the ground to be considered.
		Vector end = unit->position;
		end.z += unit->boundingBox.mins.z;
		end.z--;
	
		bsp->traceRay(unit->position, end);
	
		// can see the target so we're not on the ground
		if (tracer->outputFraction == 1.0)
		{
			return;
		}
	}
	
	if (unitIsAtEdge(unit))
	{
		unit->velocity.x = 0;
		unit->velocity.y = 0;
		unit->velocity.z = 0;
		unit->thinkTime = 0;
	}
}

bool unitCanSeeTarget(Entity *observer, Entity *target, Vector startPosition)
{
	if (!bsp->canSightTarget(startPosition, target->position))
	{
		return false;
	}
	
	if (!bbManager->canSightTarget(observer, target))
	{
		return false;
	}
	
	return true;
}

void teleportUnit()
{
	Math::removeBit(&self->flags, EF_TELEPORTING);
	self->position = self->destination;
	addTeleportParticles(self->position);
	audio->playSound(SND_TELEPORT3, (self == player) ? CH_PLAYER : CH_SPAWN);
	
	if (self != player)
	{
		self->action = self->walk;
		camera->contentType = CONTENTS_SOLID;
	}
	else
	{
		self->action = NULL;
	}
}

void faceLocation(Unit *unit, Vector location)
{
	float dx = unit->position.x - location.x;
	float dy = unit->position.y - location.y;
	
	float angle;
	 
	angle = atan2(dy, dx);
	unit->rotation.x = -90 + toDegrees(angle);
	
	//angle = atan2(dx, dz);
	//unit->rotation.z = toDegrees(angle);
}

void faceTarget(Unit *unit)
{
	if (unit->target == NULL)
	{
		return;
	}
	
	faceLocation(unit, unit->target->position);
}

char *getUnitStatus(Unit *unit)
{
	float healthPercent = unit->health;
	healthPercent /= unit->maxHealth;
	healthPercent *= 100;
	
	if (healthPercent <= 0)
	{
		return localeManager->translate("STATUS_KIA");
	}
	else if (healthPercent <= 50)
	{
		return localeManager->translate("STATUS_WOUNDED");
	}
	else
	{
		return localeManager->translate("STATUS_OK");
	}
}
